/*
Copyright  2004-2005 Eland Systems All Rights Reserved.

   1. Redistribution and use in source and binary forms must retain the above
   copyright notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

   3. The name of Eland Systems may not be used to endorse or promote products
   derived from this software without specific prior written permission.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sysexits.h>
#include <syslog.h>
#include <pthread.h>

#ifndef _LINUX
#include <sys/queue.h>
#else
#include "queue.h"
#endif

#include "lock.h"

#ifdef POPAUTH
# include <db.h>
# include <fcntl.h>
#endif

#define SCAMCONF "/etc/mail/scam.conf"

struct entry {
	char	*c;
	SLIST_ENTRY(entry)	entries;
};

SLIST_HEAD(, entry) gwhitelist;

pthread_rwlock_t gwhite_lock;

struct CIDR {
	unsigned long ip;
	short int masklen;
	SLIST_ENTRY(CIDR)	cidrs;
};

SLIST_HEAD(, CIDR)	cidrlist;

size_t
strlcpy(char *dst, const char *src, size_t siz)
{
	register char *d = dst;
	register const char *s = src;
	register size_t n = siz;

	/* Copy as many bytes as will fit */
	if (n != 0 && --n != 0) {
		do {
			if ((*d++ = *s++) == 0)
				break;
		} while (--n != 0);
	}

	/* Not enough room in dst, add NUL and traverse rest of src */
	if (n == 0) {
		if (siz != 0)
			*d = '\0';		/* NUL-terminate dst */
		while (*s++)
			;
	}

	return(s - src - 1);	/* count does not include NUL */
}

void gwhite_init()
{
	int ret;

	if ((ret = pthread_rwlock_init(&gwhite_lock, NULL)) != 0)
	{
		syslog(LOG_ERR,  "pthread_rwlock_init failed: %s", strerror(ret));
		exit(EX_OSERR);
	}
	SLIST_INIT(&cidrlist);
}

int gwhite_readlist()
{
	FILE *fh;
	struct entry *ent;
	const LINELENGTH = 1024;
	char line[LINELENGTH + 1];
	char buf[LINELENGTH + 1];
	struct CIDR *entcidr;

	WRLOCK(gwhite_lock);
	while (!SLIST_EMPTY(&gwhitelist)) {
		ent = SLIST_FIRST(&gwhitelist);
		free(ent->c);
		SLIST_REMOVE_HEAD(&gwhitelist, entries);
		free(ent);
	}

	while (!SLIST_EMPTY(&cidrlist)) {
		entcidr = SLIST_FIRST(&cidrlist);
		SLIST_REMOVE_HEAD(&cidrlist, cidrs);
		free(entcidr);
	}

	if ((fh = fopen( SCAMCONF, "r")) == NULL) {
		UNLOCK(gwhite_lock);
		return 1;
	}

	syslog (LOG_INFO, "reading greylist");

	while( fgets(line, LINELENGTH, fh) )
	{
		if (line[0] == '#')
			continue;

		if (strlen(line) > 3)
		{
			unsigned int ip1 = 0;
			unsigned int ip2 = 0;
			unsigned int ip3 = 0;
			unsigned int ip4 = 0;
			unsigned int ipm = 0;

			if (sscanf( line, "GreylistConnect:%u.%u.%u.%u/%u", &ip1, &ip2, &ip3, &ip4, &ipm) == 5)
			{
				snprintf( buf, sizeof(buf), "%u.%u.%u.%u", ip1, ip2, ip3, ip4);
				entcidr = (struct CIDR *)malloc(sizeof(struct CIDR));
				entcidr->ip = inet_addr(buf);
				entcidr->masklen = ipm;
				SLIST_INSERT_HEAD(&cidrlist, entcidr, cidrs);
			}
			else if (sscanf( line, "GreylistConnect:%u.%u.%u.%u", &ip1, &ip2, &ip3, &ip4) == 4)
			{
				snprintf( buf, sizeof(buf), "%u.%u.%u.%u", ip1, ip2, ip3, ip4);
				entcidr = (struct CIDR *)malloc(sizeof(struct CIDR));
				entcidr->ip = inet_addr(buf);
				entcidr->masklen = 32;
				SLIST_INSERT_HEAD(&cidrlist, entcidr, cidrs);
			}
			else if (sscanf( line, "GreylistConnect:%1024[A-Za-z0-9-.]", buf) == 1)
			{
				ent = (struct entry *)malloc(sizeof(struct entry));
				if ((ent->c = strdup(buf)) == NULL)
				{
					fclose(fh);
					UNLOCK(gwhite_lock);
					return 1;
				}
				SLIST_INSERT_HEAD(&gwhitelist, ent, entries);
			}
		}
	}
	fclose(fh);
	UNLOCK(gwhite_lock);
	return 0;
}

int
ip_cidr(const unsigned long ip, const short int masklen, const unsigned long checkip)
{
	unsigned long ipaddr = 0;
	unsigned long cidrip = 0;
	unsigned long subnet = 0;

	subnet = ~0;
	subnet = subnet << (32 - masklen);

	cidrip = htonl(ip) & subnet;
	ipaddr = ntohl(checkip) & subnet;

	if (cidrip == ipaddr)
		return 1;

	return 0;
}

int
check_ip(const unsigned long checkip)
{
	int ret = 0;
	struct CIDR *entcidr;

	RDLOCK(gwhite_lock);
	SLIST_FOREACH(entcidr, &cidrlist, cidrs)
	{
		if (ip_cidr(entcidr->ip, entcidr->masklen, checkip) == 1)
		{
			syslog (LOG_INFO, "greylist skip cidr");
			ret = 1;
			break;
		}
	}
	UNLOCK(gwhite_lock);

	return ret;
}


int
unixconn(int sockfd, const char *ctlsock_path)
{
	struct sockaddr_un sa;
	int rc;
	int timeout = 1500;

	memset(&sa, 0, sizeof(sa));
	sa.sun_family = AF_UNIX;
	strlcpy(sa.sun_path, ctlsock_path, sizeof(sa.sun_path));

	rc = connect(sockfd, (struct sockaddr *)&sa, sizeof(sa));

	if(-1 == rc)
	{
		switch (errno)
		{
			case EINPROGRESS:
			case EWOULDBLOCK:
			case EINTR:
				rc = waitconnect(sockfd, timeout);
				break;

			case ECONNREFUSED:
			default:
				syslog(LOG_ERR, "no server on %s\n", ctlsock_path);
				return -1;
		}
	}

	if (0 == rc)
	{
		int len = 0;
		int slen = sizeof(len);
		if ( -1 == getsockopt(sockfd, SOL_SOCKET, SO_ERROR, (void*)&len, &slen))
		{
			int err = errno;
			if ((0 == err) || (EISCONN == err))
			{
				rc = 0;
			} else
				rc = -1;
		}
	}

	return rc;
}

int
waitconnect(int sockfd, int timeout_msec)
{
  fd_set fd;
  fd_set errfd;
  struct timeval interval;
  int rc;

  /* now select() until we get connect or timeout */
  FD_ZERO(&fd);
  FD_SET(sockfd, &fd);

  FD_ZERO(&errfd);
  FD_SET(sockfd, &errfd);

  interval.tv_sec = timeout_msec/1000;
  timeout_msec -= interval.tv_sec*1000;

  interval.tv_usec = timeout_msec*1000;

  rc = select(sockfd+1, NULL, &fd, &errfd, &interval);
  if(-1 == rc)
    /* error, no connect here, try next */
    return -1;

  else if(0 == rc)
    /* timeout, no connect today */
    return 1;

  if(FD_ISSET(sockfd, &errfd)) {
    /* error condition caught */
    return 2;
  }

  /* we have a connect! */
  return 0;
}

int
dothost(char *host)
{
	int j = strlen(host);
	int i = j;
	int dot = 0;

	if (j < 5)
		return 0;

	while (i > 0)
	{
		if (host[i] == '.')
		{
			dot++;
			if (dot > 2)
			{
				return (dot);
			}
		}
		i--;
	}

	return (dot);
}

#ifdef POPAUTH

int
checkpopauth(char *dbfile, struct sockaddr *ip)
{
	int ret = 0;
	int c;
	DB *db;
	struct sockaddr_in *sin;
	struct in_addr addr;
	struct in_addr mask;
	DBT d;
	DBT q;
	char ipbuf[17];

	/* skip anything not IPv4 (for now) */
	if (ip->sa_family != AF_INET)
		return 0;
	else
		sin = (struct sockaddr_in *)ip;

	memset(&d, 0, sizeof d);
	memset(&q, 0, sizeof q);

	memcpy(&addr.s_addr, &sin->sin_addr, sizeof addr.s_addr);

	inet_ntop(AF_INET, &addr, ipbuf, sizeof ipbuf);

	q.data = ipbuf;
	q.size = strlen(q.data);

#if DB_VERSION_MAJOR > 2
	if (db_create(&db, NULL, 0) != 0 ||
	    db->open(db, dbfile, NULL, DB_HASH, DB_RDONLY, 0) != 0)
#elif DB_VERSION_MAJOR == 2
	if (db_open(dbfile, DB_HASH, DB_RDONLY, DB_MODE, NULL, NULL, &db) != 0)
#else /* DB_VERSION_MAJOR == 2 */
	if ((db = db_open(dbfile, O_RDONLY, DB_MODE, DB_HASH, NULL)) == NULL)
#endif /* DB_VERSION_MAJOR < 2 */
		return 0;

#if DB_VERSION_MAJOR < 2
	if (!((db->get) (db, &q, &d, 0)))
#else /* DB_VERSION_MAJOR < 2 */
	if (!((db->get) (db, NULL, &q, &d, 0)))
#endif
		ret = 1;

#if DB_VERSION_MAJOR < 2
	(void) db->close(db);
#else /* DB_VERSION_MAJOR < 2 */
	(void) db->close(db, 0);
#endif

	return ret;
}
#endif

